# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tidepoolTemplate_big.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1595, 810)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding
        )
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(MainWindow.sizePolicy().hasHeightForWidth())
        MainWindow.setSizePolicy(sizePolicy)
        MainWindow.setMinimumSize(QtCore.QSize(1595, 810))
        MainWindow.setMaximumSize(QtCore.QSize(1595, 810))
        MainWindow.setSizeIncrement(QtCore.QSize(1, 1))
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.imageData_groupBox = QtWidgets.QGroupBox(self.centralwidget)
        self.imageData_groupBox.setGeometry(QtCore.QRect(440, 7, 561, 591))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Maximum
        )
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.imageData_groupBox.sizePolicy().hasHeightForWidth())
        self.imageData_groupBox.setSizePolicy(sizePolicy)
        self.imageData_groupBox.setMinimumSize(QtCore.QSize(550, 156))
        self.imageData_groupBox.setMaximumSize(QtCore.QSize(1100, 1000))
        self.imageData_groupBox.setSizeIncrement(QtCore.QSize(1, 1))
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        self.imageData_groupBox.setFont(font)
        self.imageData_groupBox.setObjectName("imageData_groupBox")
        self.gridLayout = QtWidgets.QGridLayout(self.imageData_groupBox)
        self.gridLayout.setObjectName("gridLayout")
        self.gridLayout_2 = QtWidgets.QGridLayout()
        self.gridLayout_2.setSizeConstraint(QtCore.Qt.QLayout.SizeConstraint.SetMaximumSize)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setSizeConstraint(QtCore.Qt.QLayout.SizeConstraint.SetFixedSize)
        self.verticalLayout_3.setContentsMargins(-1, 5, -1, 5)
        self.verticalLayout_3.setSpacing(0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.dispmax_doubleSpinBox = QtWidgets.QDoubleSpinBox(self.imageData_groupBox)
        self.dispmax_doubleSpinBox.setAlignment(
            QtCore.Qt.Qt.AlignmentFlag.AlignRight
            | QtCore.Qt.Qt.AlignmentFlag.AlignTrailing
            | QtCore.Qt.Qt.AlignmentFlag.AlignVCenter
        )
        self.dispmax_doubleSpinBox.setDecimals(3)
        self.dispmax_doubleSpinBox.setObjectName("dispmax_doubleSpinBox")
        self.verticalLayout_3.addWidget(self.dispmax_doubleSpinBox)
        self.resetDispLimits_Button = QtWidgets.QPushButton(self.imageData_groupBox)
        self.resetDispLimits_Button.setObjectName("resetDispLimits_Button")
        self.verticalLayout_3.addWidget(self.resetDispLimits_Button)
        self.resetDispSmart_Button = QtWidgets.QPushButton(self.imageData_groupBox)
        self.resetDispSmart_Button.setObjectName("resetDispSmart_Button")
        self.verticalLayout_3.addWidget(self.resetDispSmart_Button)
        self.saveDisp_Button = QtWidgets.QPushButton(self.imageData_groupBox)
        self.saveDisp_Button.setObjectName("saveDisp_Button")
        self.verticalLayout_3.addWidget(self.saveDisp_Button)
        self.setMask_Button = QtWidgets.QPushButton(self.imageData_groupBox)
        self.setMask_Button.setObjectName("setMask_Button")
        self.verticalLayout_3.addWidget(self.setMask_Button)
        self.dispmin_doubleSpinBox = QtWidgets.QDoubleSpinBox(self.imageData_groupBox)
        self.dispmin_doubleSpinBox.setAlignment(
            QtCore.Qt.Qt.AlignmentFlag.AlignRight
            | QtCore.Qt.Qt.AlignmentFlag.AlignTrailing
            | QtCore.Qt.Qt.AlignmentFlag.AlignVCenter
        )
        self.dispmin_doubleSpinBox.setDecimals(3)
        self.dispmin_doubleSpinBox.setObjectName("dispmin_doubleSpinBox")
        self.verticalLayout_3.addWidget(self.dispmin_doubleSpinBox)
        self.transparency_checkBox = QtWidgets.QCheckBox(self.imageData_groupBox)
        self.transparency_checkBox.setChecked(False)
        self.transparency_checkBox.setObjectName("transparency_checkBox")
        self.verticalLayout_3.addWidget(self.transparency_checkBox)
        self.gridLayout_2.addLayout(self.verticalLayout_3, 0, 1, 1, 1)
        self.graphicsView_colorbar = GraphicsView(self.imageData_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.graphicsView_colorbar.sizePolicy().hasHeightForWidth())
        self.graphicsView_colorbar.setSizePolicy(sizePolicy)
        self.graphicsView_colorbar.setMinimumSize(QtCore.QSize(25, 128))
        self.graphicsView_colorbar.setMaximumSize(QtCore.QSize(25, 256))
        self.graphicsView_colorbar.setSizeIncrement(QtCore.QSize(0, 1))
        self.graphicsView_colorbar.setResizeAnchor(
            QtCore.Qt.QGraphicsView.ViewportAnchor.AnchorViewCenter
        )
        self.graphicsView_colorbar.setObjectName("graphicsView_colorbar")
        self.gridLayout_2.addWidget(self.graphicsView_colorbar, 0, 0, 1, 1)
        self.aspects_verticalLayout = QtWidgets.QVBoxLayout()
        self.aspects_verticalLayout.setSizeConstraint(
            QtCore.Qt.QLayout.SizeConstraint.SetMinimumSize
        )
        self.aspects_verticalLayout.setSpacing(0)
        self.aspects_verticalLayout.setObjectName("aspects_verticalLayout")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout()
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.label_2 = QtWidgets.QLabel(self.imageData_groupBox)
        self.label_2.setMaximumSize(QtCore.QSize(16777215, 14))
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        self.label_2.setFont(font)
        self.label_2.setObjectName("label_2")
        self.verticalLayout_4.addWidget(self.label_2)
        self.gray_radioButton = QtWidgets.QRadioButton(self.imageData_groupBox)
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        font.setPointSize(10)
        self.gray_radioButton.setFont(font)
        self.gray_radioButton.setObjectName("gray_radioButton")
        self.colormap_buttonGroup = QtWidgets.QButtonGroup(MainWindow)
        self.colormap_buttonGroup.setObjectName("colormap_buttonGroup")
        self.colormap_buttonGroup.addButton(self.gray_radioButton)
        self.verticalLayout_4.addWidget(self.gray_radioButton)
        self.thermal_radioButton = QtWidgets.QRadioButton(self.imageData_groupBox)
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        font.setPointSize(10)
        self.thermal_radioButton.setFont(font)
        self.thermal_radioButton.setObjectName("thermal_radioButton")
        self.colormap_buttonGroup.addButton(self.thermal_radioButton)
        self.verticalLayout_4.addWidget(self.thermal_radioButton)
        self.plasma_radioButton = QtWidgets.QRadioButton(self.imageData_groupBox)
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        font.setPointSize(10)
        self.plasma_radioButton.setFont(font)
        self.plasma_radioButton.setObjectName("plasma_radioButton")
        self.colormap_buttonGroup.addButton(self.plasma_radioButton)
        self.verticalLayout_4.addWidget(self.plasma_radioButton)
        self.rainbow_radioButton = QtWidgets.QRadioButton(self.imageData_groupBox)
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        font.setPointSize(10)
        self.rainbow_radioButton.setFont(font)
        self.rainbow_radioButton.setObjectName("rainbow_radioButton")
        self.colormap_buttonGroup.addButton(self.rainbow_radioButton)
        self.verticalLayout_4.addWidget(self.rainbow_radioButton)
        self.viridis_radioButton = QtWidgets.QRadioButton(self.imageData_groupBox)
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        font.setPointSize(10)
        self.viridis_radioButton.setFont(font)
        self.viridis_radioButton.setObjectName("viridis_radioButton")
        self.colormap_buttonGroup.addButton(self.viridis_radioButton)
        self.verticalLayout_4.addWidget(self.viridis_radioButton)
        self.turbo_radioButton = QtWidgets.QRadioButton(self.imageData_groupBox)
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        font.setPointSize(10)
        self.turbo_radioButton.setFont(font)
        self.turbo_radioButton.setObjectName("turbo_radioButton")
        self.colormap_buttonGroup.addButton(self.turbo_radioButton)
        self.verticalLayout_4.addWidget(self.turbo_radioButton)
        self.aspects_verticalLayout.addLayout(self.verticalLayout_4)
        self.summary_Label = QtWidgets.QLabel(self.imageData_groupBox)
        self.summary_Label.setMaximumSize(QtCore.QSize(16777215, 14))
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        self.summary_Label.setFont(font)
        self.summary_Label.setObjectName("summary_Label")
        self.aspects_verticalLayout.addWidget(self.summary_Label)
        self.raw_radioButton = QtWidgets.QRadioButton(self.imageData_groupBox)
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        font.setPointSize(10)
        self.raw_radioButton.setFont(font)
        self.raw_radioButton.setChecked(True)
        self.raw_radioButton.setObjectName("raw_radioButton")
        self.stats_buttonGroup = QtWidgets.QButtonGroup(MainWindow)
        self.stats_buttonGroup.setObjectName("stats_buttonGroup")
        self.stats_buttonGroup.addButton(self.raw_radioButton)
        self.aspects_verticalLayout.addWidget(self.raw_radioButton)
        self.mean_radioButton = QtWidgets.QRadioButton(self.imageData_groupBox)
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        font.setPointSize(10)
        self.mean_radioButton.setFont(font)
        self.mean_radioButton.setChecked(False)
        self.mean_radioButton.setObjectName("mean_radioButton")
        self.stats_buttonGroup.addButton(self.mean_radioButton)
        self.aspects_verticalLayout.addWidget(self.mean_radioButton)
        self.median_radioButton = QtWidgets.QRadioButton(self.imageData_groupBox)
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        font.setPointSize(10)
        self.median_radioButton.setFont(font)
        self.median_radioButton.setObjectName("median_radioButton")
        self.stats_buttonGroup.addButton(self.median_radioButton)
        self.aspects_verticalLayout.addWidget(self.median_radioButton)
        self.CoV_radioButton = QtWidgets.QRadioButton(self.imageData_groupBox)
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        font.setPointSize(10)
        self.CoV_radioButton.setFont(font)
        self.CoV_radioButton.setObjectName("CoV_radioButton")
        self.stats_buttonGroup.addButton(self.CoV_radioButton)
        self.aspects_verticalLayout.addWidget(self.CoV_radioButton)
        self.std_radioButton = QtWidgets.QRadioButton(self.imageData_groupBox)
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        font.setPointSize(10)
        self.std_radioButton.setFont(font)
        self.std_radioButton.setObjectName("std_radioButton")
        self.stats_buttonGroup.addButton(self.std_radioButton)
        self.aspects_verticalLayout.addWidget(self.std_radioButton)
        self.MAD_radioButton = QtWidgets.QRadioButton(self.imageData_groupBox)
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        font.setPointSize(10)
        self.MAD_radioButton.setFont(font)
        self.MAD_radioButton.setObjectName("MAD_radioButton")
        self.stats_buttonGroup.addButton(self.MAD_radioButton)
        self.aspects_verticalLayout.addWidget(self.MAD_radioButton)
        self.gridLayout_2.addLayout(self.aspects_verticalLayout, 0, 2, 1, 1)
        self.gridLayout.addLayout(self.gridLayout_2, 2, 2, 1, 1)
        self.main_graphicsView_cor = GraphicsLayoutWidget(self.imageData_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.main_graphicsView_cor.sizePolicy().hasHeightForWidth())
        self.main_graphicsView_cor.setSizePolicy(sizePolicy)
        self.main_graphicsView_cor.setMinimumSize(QtCore.QSize(256, 128))
        self.main_graphicsView_cor.setMaximumSize(QtCore.QSize(256, 256))
        self.main_graphicsView_cor.setSizeIncrement(QtCore.QSize(1, 1))
        self.main_graphicsView_cor.setFrameShape(QtCore.Qt.QFrame.Shape.NoFrame)
        self.main_graphicsView_cor.setAlignment(QtCore.Qt.Qt.AlignmentFlag.AlignCenter)
        self.main_graphicsView_cor.setObjectName("main_graphicsView_cor")
        self.gridLayout.addWidget(self.main_graphicsView_cor, 1, 2, 1, 1)
        self.main_graphicsView_sag = GraphicsLayoutWidget(self.imageData_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.main_graphicsView_sag.sizePolicy().hasHeightForWidth())
        self.main_graphicsView_sag.setSizePolicy(sizePolicy)
        self.main_graphicsView_sag.setMinimumSize(QtCore.QSize(256, 256))
        self.main_graphicsView_sag.setMaximumSize(QtCore.QSize(256, 256))
        self.main_graphicsView_sag.setSizeIncrement(QtCore.QSize(1, 1))
        self.main_graphicsView_sag.setFrameShape(QtCore.Qt.QFrame.Shape.NoFrame)
        self.main_graphicsView_sag.setAlignment(QtCore.Qt.Qt.AlignmentFlag.AlignCenter)
        self.main_graphicsView_sag.setObjectName("main_graphicsView_sag")
        self.gridLayout.addWidget(self.main_graphicsView_sag, 2, 1, 1, 1)
        self.main_graphicsView_ax = GraphicsLayoutWidget(self.imageData_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.main_graphicsView_ax.sizePolicy().hasHeightForWidth())
        self.main_graphicsView_ax.setSizePolicy(sizePolicy)
        self.main_graphicsView_ax.setMinimumSize(QtCore.QSize(256, 256))
        self.main_graphicsView_ax.setMaximumSize(QtCore.QSize(256, 256))
        self.main_graphicsView_ax.setSizeIncrement(QtCore.QSize(1, 1))
        self.main_graphicsView_ax.setFrameShape(QtCore.Qt.QFrame.Shape.NoFrame)
        self.main_graphicsView_ax.setAlignment(QtCore.Qt.Qt.AlignmentFlag.AlignCenter)
        self.main_graphicsView_ax.setViewportUpdateMode(
            QtCore.Qt.QGraphicsView.ViewportUpdateMode.SmartViewportUpdate
        )
        self.main_graphicsView_ax.setObjectName("main_graphicsView_ax")
        self.gridLayout.addWidget(self.main_graphicsView_ax, 1, 1, 1, 1)
        self.overlays_groupBox = QtWidgets.QGroupBox(self.centralwidget)
        self.overlays_groupBox.setGeometry(QtCore.QRect(10, 7, 421, 591))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.overlays_groupBox.sizePolicy().hasHeightForWidth())
        self.overlays_groupBox.setSizePolicy(sizePolicy)
        self.overlays_groupBox.setMinimumSize(QtCore.QSize(210, 0))
        self.overlays_groupBox.setMaximumSize(QtCore.QSize(16777215, 683))
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        self.overlays_groupBox.setFont(font)
        self.overlays_groupBox.setObjectName("overlays_groupBox")
        self.overlay_radioButton_05 = QtWidgets.QRadioButton(self.overlays_groupBox)
        self.overlay_radioButton_05.setGeometry(QtCore.QRect(10, 351, 151, 20))
        font = QtGui.QFont()
        font.setFamily(".AppleSystemUIFont")
        font.setPointSize(10)
        self.overlay_radioButton_05.setFont(font)
        self.overlay_radioButton_05.setStyleSheet("color: rgb(255, 255, 255);")
        self.overlay_radioButton_05.setObjectName("overlay_radioButton_05")
        self.overlay_graphicsView_01 = GraphicsLayoutWidget(self.overlays_groupBox)
        self.overlay_graphicsView_01.setGeometry(QtCore.QRect(10, 25, 192, 64))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.overlay_graphicsView_01.sizePolicy().hasHeightForWidth())
        self.overlay_graphicsView_01.setSizePolicy(sizePolicy)
        self.overlay_graphicsView_01.setMinimumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_01.setMaximumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_01.setFrameShape(QtCore.Qt.QFrame.Shape.NoFrame)
        self.overlay_graphicsView_01.setAlignment(QtCore.Qt.Qt.AlignmentFlag.AlignCenter)
        self.overlay_graphicsView_01.setObjectName("overlay_graphicsView_01")
        self.overlay_radioButton_01 = QtWidgets.QRadioButton(self.overlays_groupBox)
        self.overlay_radioButton_01.setGeometry(QtCore.QRect(10, 71, 151, 20))
        font = QtGui.QFont()
        font.setFamily(".AppleSystemUIFont")
        font.setPointSize(10)
        self.overlay_radioButton_01.setFont(font)
        self.overlay_radioButton_01.setAutoFillBackground(False)
        self.overlay_radioButton_01.setStyleSheet("color: rgb(255, 255, 255);")
        self.overlay_radioButton_01.setObjectName("overlay_radioButton_01")
        self.overlay_radioButton_02 = QtWidgets.QRadioButton(self.overlays_groupBox)
        self.overlay_radioButton_02.setGeometry(QtCore.QRect(10, 141, 151, 20))
        font = QtGui.QFont()
        font.setFamily(".AppleSystemUIFont")
        font.setPointSize(10)
        self.overlay_radioButton_02.setFont(font)
        self.overlay_radioButton_02.setStyleSheet("color: rgb(255, 255, 255);")
        self.overlay_radioButton_02.setObjectName("overlay_radioButton_02")
        self.overlay_radioButton_03 = QtWidgets.QRadioButton(self.overlays_groupBox)
        self.overlay_radioButton_03.setGeometry(QtCore.QRect(10, 211, 151, 20))
        font = QtGui.QFont()
        font.setFamily(".AppleSystemUIFont")
        font.setPointSize(10)
        self.overlay_radioButton_03.setFont(font)
        self.overlay_radioButton_03.setStyleSheet("color: rgb(255, 255, 255);")
        self.overlay_radioButton_03.setObjectName("overlay_radioButton_03")
        self.overlay_radioButton_04 = QtWidgets.QRadioButton(self.overlays_groupBox)
        self.overlay_radioButton_04.setGeometry(QtCore.QRect(10, 281, 151, 20))
        font = QtGui.QFont()
        font.setFamily(".AppleSystemUIFont")
        font.setPointSize(10)
        self.overlay_radioButton_04.setFont(font)
        self.overlay_radioButton_04.setStyleSheet("color: rgb(255, 255, 255);")
        self.overlay_radioButton_04.setObjectName("overlay_radioButton_04")
        self.corrout_radioButton = QtWidgets.QRadioButton(self.overlays_groupBox)
        self.corrout_radioButton.setGeometry(QtCore.QRect(10, 620, 151, 20))
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        font.setPointSize(10)
        self.corrout_radioButton.setFont(font)
        self.corrout_radioButton.setChecked(True)
        self.corrout_radioButton.setObjectName("corrout_radioButton")
        self.overlay_radioButton_06 = QtWidgets.QRadioButton(self.overlays_groupBox)
        self.overlay_radioButton_06.setGeometry(QtCore.QRect(10, 421, 151, 20))
        font = QtGui.QFont()
        font.setFamily(".AppleSystemUIFont")
        font.setPointSize(10)
        self.overlay_radioButton_06.setFont(font)
        self.overlay_radioButton_06.setStyleSheet("color: rgb(255, 255, 255);")
        self.overlay_radioButton_06.setObjectName("overlay_radioButton_06")
        self.overlay_radioButton_07 = QtWidgets.QRadioButton(self.overlays_groupBox)
        self.overlay_radioButton_07.setGeometry(QtCore.QRect(10, 491, 151, 20))
        font = QtGui.QFont()
        font.setFamily(".AppleSystemUIFont")
        font.setPointSize(10)
        self.overlay_radioButton_07.setFont(font)
        self.overlay_radioButton_07.setStyleSheet("color: rgb(255, 255, 255);")
        self.overlay_radioButton_07.setObjectName("overlay_radioButton_07")
        self.overlay_radioButton_08 = QtWidgets.QRadioButton(self.overlays_groupBox)
        self.overlay_radioButton_08.setGeometry(QtCore.QRect(10, 561, 151, 20))
        font = QtGui.QFont()
        font.setFamily(".AppleSystemUIFont")
        font.setPointSize(10)
        self.overlay_radioButton_08.setFont(font)
        self.overlay_radioButton_08.setStyleSheet("color: rgb(255, 255, 255);")
        self.overlay_radioButton_08.setObjectName("overlay_radioButton_08")
        self.overlay_graphicsView_04 = GraphicsLayoutWidget(self.overlays_groupBox)
        self.overlay_graphicsView_04.setGeometry(QtCore.QRect(10, 235, 192, 64))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.overlay_graphicsView_04.sizePolicy().hasHeightForWidth())
        self.overlay_graphicsView_04.setSizePolicy(sizePolicy)
        self.overlay_graphicsView_04.setMinimumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_04.setMaximumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_04.setFrameShape(QtCore.Qt.QFrame.Shape.NoFrame)
        self.overlay_graphicsView_04.setAlignment(QtCore.Qt.Qt.AlignmentFlag.AlignCenter)
        self.overlay_graphicsView_04.setObjectName("overlay_graphicsView_04")
        self.overlay_graphicsView_06 = GraphicsLayoutWidget(self.overlays_groupBox)
        self.overlay_graphicsView_06.setGeometry(QtCore.QRect(10, 375, 192, 64))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.overlay_graphicsView_06.sizePolicy().hasHeightForWidth())
        self.overlay_graphicsView_06.setSizePolicy(sizePolicy)
        self.overlay_graphicsView_06.setMinimumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_06.setMaximumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_06.setFrameShape(QtCore.Qt.QFrame.Shape.NoFrame)
        self.overlay_graphicsView_06.setAlignment(QtCore.Qt.Qt.AlignmentFlag.AlignCenter)
        self.overlay_graphicsView_06.setObjectName("overlay_graphicsView_06")
        self.overlay_graphicsView_03 = GraphicsLayoutWidget(self.overlays_groupBox)
        self.overlay_graphicsView_03.setGeometry(QtCore.QRect(10, 165, 192, 64))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.overlay_graphicsView_03.sizePolicy().hasHeightForWidth())
        self.overlay_graphicsView_03.setSizePolicy(sizePolicy)
        self.overlay_graphicsView_03.setMinimumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_03.setMaximumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_03.setFrameShape(QtCore.Qt.QFrame.Shape.NoFrame)
        self.overlay_graphicsView_03.setAlignment(QtCore.Qt.Qt.AlignmentFlag.AlignCenter)
        self.overlay_graphicsView_03.setObjectName("overlay_graphicsView_03")
        self.overlay_graphicsView_08 = GraphicsLayoutWidget(self.overlays_groupBox)
        self.overlay_graphicsView_08.setGeometry(QtCore.QRect(10, 515, 192, 64))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.overlay_graphicsView_08.sizePolicy().hasHeightForWidth())
        self.overlay_graphicsView_08.setSizePolicy(sizePolicy)
        self.overlay_graphicsView_08.setMinimumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_08.setMaximumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_08.setFrameShape(QtCore.Qt.QFrame.Shape.NoFrame)
        self.overlay_graphicsView_08.setAlignment(QtCore.Qt.Qt.AlignmentFlag.AlignCenter)
        self.overlay_graphicsView_08.setObjectName("overlay_graphicsView_08")
        self.overlay_graphicsView_05 = GraphicsLayoutWidget(self.overlays_groupBox)
        self.overlay_graphicsView_05.setGeometry(QtCore.QRect(10, 305, 192, 64))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.overlay_graphicsView_05.sizePolicy().hasHeightForWidth())
        self.overlay_graphicsView_05.setSizePolicy(sizePolicy)
        self.overlay_graphicsView_05.setMinimumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_05.setMaximumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_05.setFrameShape(QtCore.Qt.QFrame.Shape.NoFrame)
        self.overlay_graphicsView_05.setAlignment(QtCore.Qt.Qt.AlignmentFlag.AlignCenter)
        self.overlay_graphicsView_05.setObjectName("overlay_graphicsView_05")
        self.overlay_graphicsView_02 = GraphicsLayoutWidget(self.overlays_groupBox)
        self.overlay_graphicsView_02.setGeometry(QtCore.QRect(10, 95, 192, 64))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.overlay_graphicsView_02.sizePolicy().hasHeightForWidth())
        self.overlay_graphicsView_02.setSizePolicy(sizePolicy)
        self.overlay_graphicsView_02.setMinimumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_02.setMaximumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_02.setFrameShape(QtCore.Qt.QFrame.Shape.NoFrame)
        self.overlay_graphicsView_02.setAlignment(QtCore.Qt.Qt.AlignmentFlag.AlignCenter)
        self.overlay_graphicsView_02.setObjectName("overlay_graphicsView_02")
        self.overlay_graphicsView_07 = GraphicsLayoutWidget(self.overlays_groupBox)
        self.overlay_graphicsView_07.setGeometry(QtCore.QRect(10, 445, 192, 64))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.overlay_graphicsView_07.sizePolicy().hasHeightForWidth())
        self.overlay_graphicsView_07.setSizePolicy(sizePolicy)
        self.overlay_graphicsView_07.setMinimumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_07.setMaximumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_07.setFrameShape(QtCore.Qt.QFrame.Shape.NoFrame)
        self.overlay_graphicsView_07.setAlignment(QtCore.Qt.Qt.AlignmentFlag.AlignCenter)
        self.overlay_graphicsView_07.setObjectName("overlay_graphicsView_07")
        self.overlay_graphicsView_09 = GraphicsLayoutWidget(self.overlays_groupBox)
        self.overlay_graphicsView_09.setGeometry(QtCore.QRect(210, 25, 192, 64))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.overlay_graphicsView_09.sizePolicy().hasHeightForWidth())
        self.overlay_graphicsView_09.setSizePolicy(sizePolicy)
        self.overlay_graphicsView_09.setMinimumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_09.setMaximumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_09.setFrameShape(QtCore.Qt.QFrame.Shape.NoFrame)
        self.overlay_graphicsView_09.setAlignment(QtCore.Qt.Qt.AlignmentFlag.AlignCenter)
        self.overlay_graphicsView_09.setObjectName("overlay_graphicsView_09")
        self.overlay_graphicsView_11 = GraphicsLayoutWidget(self.overlays_groupBox)
        self.overlay_graphicsView_11.setGeometry(QtCore.QRect(210, 165, 192, 64))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.overlay_graphicsView_11.sizePolicy().hasHeightForWidth())
        self.overlay_graphicsView_11.setSizePolicy(sizePolicy)
        self.overlay_graphicsView_11.setMinimumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_11.setMaximumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_11.setFrameShape(QtCore.Qt.QFrame.Shape.NoFrame)
        self.overlay_graphicsView_11.setAlignment(QtCore.Qt.Qt.AlignmentFlag.AlignCenter)
        self.overlay_graphicsView_11.setObjectName("overlay_graphicsView_11")
        self.overlay_graphicsView_10 = GraphicsLayoutWidget(self.overlays_groupBox)
        self.overlay_graphicsView_10.setGeometry(QtCore.QRect(210, 95, 192, 64))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.overlay_graphicsView_10.sizePolicy().hasHeightForWidth())
        self.overlay_graphicsView_10.setSizePolicy(sizePolicy)
        self.overlay_graphicsView_10.setMinimumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_10.setMaximumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_10.setFrameShape(QtCore.Qt.QFrame.Shape.NoFrame)
        self.overlay_graphicsView_10.setAlignment(QtCore.Qt.Qt.AlignmentFlag.AlignCenter)
        self.overlay_graphicsView_10.setObjectName("overlay_graphicsView_10")
        self.overlay_graphicsView_12 = GraphicsLayoutWidget(self.overlays_groupBox)
        self.overlay_graphicsView_12.setGeometry(QtCore.QRect(210, 235, 192, 64))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.overlay_graphicsView_12.sizePolicy().hasHeightForWidth())
        self.overlay_graphicsView_12.setSizePolicy(sizePolicy)
        self.overlay_graphicsView_12.setMinimumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_12.setMaximumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_12.setFrameShape(QtCore.Qt.QFrame.Shape.NoFrame)
        self.overlay_graphicsView_12.setAlignment(QtCore.Qt.Qt.AlignmentFlag.AlignCenter)
        self.overlay_graphicsView_12.setObjectName("overlay_graphicsView_12")
        self.overlay_graphicsView_13 = GraphicsLayoutWidget(self.overlays_groupBox)
        self.overlay_graphicsView_13.setGeometry(QtCore.QRect(210, 305, 192, 64))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.overlay_graphicsView_13.sizePolicy().hasHeightForWidth())
        self.overlay_graphicsView_13.setSizePolicy(sizePolicy)
        self.overlay_graphicsView_13.setMinimumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_13.setMaximumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_13.setFrameShape(QtCore.Qt.QFrame.Shape.NoFrame)
        self.overlay_graphicsView_13.setAlignment(QtCore.Qt.Qt.AlignmentFlag.AlignCenter)
        self.overlay_graphicsView_13.setObjectName("overlay_graphicsView_13")
        self.overlay_graphicsView_14 = GraphicsLayoutWidget(self.overlays_groupBox)
        self.overlay_graphicsView_14.setGeometry(QtCore.QRect(210, 375, 192, 64))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.overlay_graphicsView_14.sizePolicy().hasHeightForWidth())
        self.overlay_graphicsView_14.setSizePolicy(sizePolicy)
        self.overlay_graphicsView_14.setMinimumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_14.setMaximumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_14.setFrameShape(QtCore.Qt.QFrame.Shape.NoFrame)
        self.overlay_graphicsView_14.setAlignment(QtCore.Qt.Qt.AlignmentFlag.AlignCenter)
        self.overlay_graphicsView_14.setObjectName("overlay_graphicsView_14")
        self.overlay_graphicsView_15 = GraphicsLayoutWidget(self.overlays_groupBox)
        self.overlay_graphicsView_15.setGeometry(QtCore.QRect(210, 445, 192, 64))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.overlay_graphicsView_15.sizePolicy().hasHeightForWidth())
        self.overlay_graphicsView_15.setSizePolicy(sizePolicy)
        self.overlay_graphicsView_15.setMinimumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_15.setMaximumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_15.setFrameShape(QtCore.Qt.QFrame.Shape.NoFrame)
        self.overlay_graphicsView_15.setAlignment(QtCore.Qt.Qt.AlignmentFlag.AlignCenter)
        self.overlay_graphicsView_15.setObjectName("overlay_graphicsView_15")
        self.overlay_graphicsView_16 = GraphicsLayoutWidget(self.overlays_groupBox)
        self.overlay_graphicsView_16.setGeometry(QtCore.QRect(210, 515, 192, 64))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.overlay_graphicsView_16.sizePolicy().hasHeightForWidth())
        self.overlay_graphicsView_16.setSizePolicy(sizePolicy)
        self.overlay_graphicsView_16.setMinimumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_16.setMaximumSize(QtCore.QSize(192, 64))
        self.overlay_graphicsView_16.setFrameShape(QtCore.Qt.QFrame.Shape.NoFrame)
        self.overlay_graphicsView_16.setAlignment(QtCore.Qt.Qt.AlignmentFlag.AlignCenter)
        self.overlay_graphicsView_16.setObjectName("overlay_graphicsView_16")
        self.overlay_radioButton_09 = QtWidgets.QRadioButton(self.overlays_groupBox)
        self.overlay_radioButton_09.setGeometry(QtCore.QRect(210, 70, 151, 20))
        font = QtGui.QFont()
        font.setFamily(".AppleSystemUIFont")
        font.setPointSize(10)
        self.overlay_radioButton_09.setFont(font)
        self.overlay_radioButton_09.setAutoFillBackground(False)
        self.overlay_radioButton_09.setStyleSheet("color: rgb(255, 255, 255);")
        self.overlay_radioButton_09.setObjectName("overlay_radioButton_09")
        self.overlay_radioButton_10 = QtWidgets.QRadioButton(self.overlays_groupBox)
        self.overlay_radioButton_10.setGeometry(QtCore.QRect(210, 140, 151, 20))
        font = QtGui.QFont()
        font.setFamily(".AppleSystemUIFont")
        font.setPointSize(10)
        self.overlay_radioButton_10.setFont(font)
        self.overlay_radioButton_10.setAutoFillBackground(False)
        self.overlay_radioButton_10.setStyleSheet("color: rgb(255, 255, 255);")
        self.overlay_radioButton_10.setObjectName("overlay_radioButton_10")
        self.overlay_radioButton_11 = QtWidgets.QRadioButton(self.overlays_groupBox)
        self.overlay_radioButton_11.setGeometry(QtCore.QRect(210, 210, 151, 20))
        font = QtGui.QFont()
        font.setFamily(".AppleSystemUIFont")
        font.setPointSize(10)
        self.overlay_radioButton_11.setFont(font)
        self.overlay_radioButton_11.setAutoFillBackground(False)
        self.overlay_radioButton_11.setStyleSheet("color: rgb(255, 255, 255);")
        self.overlay_radioButton_11.setObjectName("overlay_radioButton_11")
        self.overlay_radioButton_12 = QtWidgets.QRadioButton(self.overlays_groupBox)
        self.overlay_radioButton_12.setGeometry(QtCore.QRect(210, 280, 151, 20))
        font = QtGui.QFont()
        font.setFamily(".AppleSystemUIFont")
        font.setPointSize(10)
        self.overlay_radioButton_12.setFont(font)
        self.overlay_radioButton_12.setAutoFillBackground(False)
        self.overlay_radioButton_12.setStyleSheet("color: rgb(255, 255, 255);")
        self.overlay_radioButton_12.setObjectName("overlay_radioButton_12")
        self.overlay_radioButton_13 = QtWidgets.QRadioButton(self.overlays_groupBox)
        self.overlay_radioButton_13.setGeometry(QtCore.QRect(210, 350, 151, 20))
        font = QtGui.QFont()
        font.setFamily(".AppleSystemUIFont")
        font.setPointSize(10)
        self.overlay_radioButton_13.setFont(font)
        self.overlay_radioButton_13.setAutoFillBackground(False)
        self.overlay_radioButton_13.setStyleSheet("color: rgb(255, 255, 255);")
        self.overlay_radioButton_13.setObjectName("overlay_radioButton_13")
        self.overlay_radioButton_14 = QtWidgets.QRadioButton(self.overlays_groupBox)
        self.overlay_radioButton_14.setGeometry(QtCore.QRect(210, 420, 151, 20))
        font = QtGui.QFont()
        font.setFamily(".AppleSystemUIFont")
        font.setPointSize(10)
        self.overlay_radioButton_14.setFont(font)
        self.overlay_radioButton_14.setAutoFillBackground(False)
        self.overlay_radioButton_14.setStyleSheet("color: rgb(255, 255, 255);")
        self.overlay_radioButton_14.setObjectName("overlay_radioButton_14")
        self.overlay_radioButton_15 = QtWidgets.QRadioButton(self.overlays_groupBox)
        self.overlay_radioButton_15.setGeometry(QtCore.QRect(210, 490, 151, 20))
        font = QtGui.QFont()
        font.setFamily(".AppleSystemUIFont")
        font.setPointSize(10)
        self.overlay_radioButton_15.setFont(font)
        self.overlay_radioButton_15.setAutoFillBackground(False)
        self.overlay_radioButton_15.setStyleSheet("color: rgb(255, 255, 255);")
        self.overlay_radioButton_15.setObjectName("overlay_radioButton_15")
        self.overlay_radioButton_16 = QtWidgets.QRadioButton(self.overlays_groupBox)
        self.overlay_radioButton_16.setGeometry(QtCore.QRect(210, 560, 151, 20))
        font = QtGui.QFont()
        font.setFamily(".AppleSystemUIFont")
        font.setPointSize(10)
        self.overlay_radioButton_16.setFont(font)
        self.overlay_radioButton_16.setAutoFillBackground(False)
        self.overlay_radioButton_16.setStyleSheet("color: rgb(255, 255, 255);")
        self.overlay_radioButton_16.setObjectName("overlay_radioButton_16")
        self.overlay_graphicsView_04.raise_()
        self.overlay_graphicsView_01.raise_()
        self.overlay_radioButton_01.raise_()
        self.overlay_radioButton_04.raise_()
        self.overlay_graphicsView_06.raise_()
        self.overlay_graphicsView_08.raise_()
        self.overlay_graphicsView_05.raise_()
        self.overlay_graphicsView_02.raise_()
        self.overlay_graphicsView_07.raise_()
        self.overlay_radioButton_02.raise_()
        self.overlay_graphicsView_03.raise_()
        self.overlay_radioButton_03.raise_()
        self.overlay_radioButton_06.raise_()
        self.overlay_radioButton_08.raise_()
        self.overlay_radioButton_07.raise_()
        self.overlay_radioButton_05.raise_()
        self.corrout_radioButton.raise_()
        self.overlay_graphicsView_09.raise_()
        self.overlay_graphicsView_11.raise_()
        self.overlay_graphicsView_10.raise_()
        self.overlay_graphicsView_12.raise_()
        self.overlay_graphicsView_13.raise_()
        self.overlay_graphicsView_14.raise_()
        self.overlay_graphicsView_15.raise_()
        self.overlay_graphicsView_16.raise_()
        self.overlay_radioButton_09.raise_()
        self.overlay_radioButton_10.raise_()
        self.overlay_radioButton_11.raise_()
        self.overlay_radioButton_12.raise_()
        self.overlay_radioButton_13.raise_()
        self.overlay_radioButton_14.raise_()
        self.overlay_radioButton_15.raise_()
        self.overlay_radioButton_16.raise_()
        self.histogram_groupBox = QtWidgets.QGroupBox(self.centralwidget)
        self.histogram_groupBox.setGeometry(QtCore.QRect(1010, 225, 571, 186))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.histogram_groupBox.sizePolicy().hasHeightForWidth())
        self.histogram_groupBox.setSizePolicy(sizePolicy)
        self.histogram_groupBox.setMinimumSize(QtCore.QSize(100, 100))
        self.histogram_groupBox.setMaximumSize(QtCore.QSize(800, 1000))
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        self.histogram_groupBox.setFont(font)
        self.histogram_groupBox.setObjectName("histogram_groupBox")
        self.histogram_graphicsView = GraphicsLayoutWidget(self.histogram_groupBox)
        self.histogram_graphicsView.setGeometry(QtCore.QRect(10, 25, 551, 151))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.MinimumExpanding
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.histogram_graphicsView.sizePolicy().hasHeightForWidth())
        self.histogram_graphicsView.setSizePolicy(sizePolicy)
        self.histogram_graphicsView.setMinimumSize(QtCore.QSize(0, 100))
        self.histogram_graphicsView.setMaximumSize(QtCore.QSize(1000, 1000))
        self.histogram_graphicsView.setObjectName("histogram_graphicsView")
        self.mainimagedata_groupBox = QtWidgets.QGroupBox(self.centralwidget)
        self.mainimagedata_groupBox.setGeometry(QtCore.QRect(1010, 7, 181, 221))
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        self.mainimagedata_groupBox.setFont(font)
        self.mainimagedata_groupBox.setObjectName("mainimagedata_groupBox")
        self.label_4 = QtWidgets.QLabel(self.mainimagedata_groupBox)
        self.label_4.setGeometry(QtCore.QRect(10, 35, 16, 16))
        self.label_4.setObjectName("label_4")
        self.pixnumX_doubleSpinBox = QtWidgets.QDoubleSpinBox(self.mainimagedata_groupBox)
        self.pixnumX_doubleSpinBox.setGeometry(QtCore.QRect(30, 30, 67, 24))
        self.pixnumX_doubleSpinBox.setAlignment(
            QtCore.Qt.Qt.AlignmentFlag.AlignRight
            | QtCore.Qt.Qt.AlignmentFlag.AlignTrailing
            | QtCore.Qt.Qt.AlignmentFlag.AlignVCenter
        )
        self.pixnumX_doubleSpinBox.setDecimals(0)
        self.pixnumX_doubleSpinBox.setObjectName("pixnumX_doubleSpinBox")
        self.label_7 = QtWidgets.QLabel(self.mainimagedata_groupBox)
        self.label_7.setGeometry(QtCore.QRect(30, 20, 56, 13))
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        self.label_7.setFont(font)
        self.label_7.setObjectName("label_7")
        self.coordX_doubleSpinBox = QtWidgets.QDoubleSpinBox(self.mainimagedata_groupBox)
        self.coordX_doubleSpinBox.setGeometry(QtCore.QRect(100, 30, 67, 24))
        self.coordX_doubleSpinBox.setAlignment(
            QtCore.Qt.Qt.AlignmentFlag.AlignRight
            | QtCore.Qt.Qt.AlignmentFlag.AlignTrailing
            | QtCore.Qt.Qt.AlignmentFlag.AlignVCenter
        )
        self.coordX_doubleSpinBox.setDecimals(3)
        self.coordX_doubleSpinBox.setObjectName("coordX_doubleSpinBox")
        self.label_9 = QtWidgets.QLabel(self.mainimagedata_groupBox)
        self.label_9.setGeometry(QtCore.QRect(100, 20, 56, 13))
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        self.label_9.setFont(font)
        self.label_9.setObjectName("label_9")
        self.pixnumY_doubleSpinBox = QtWidgets.QDoubleSpinBox(self.mainimagedata_groupBox)
        self.pixnumY_doubleSpinBox.setGeometry(QtCore.QRect(30, 55, 67, 24))
        self.pixnumY_doubleSpinBox.setAlignment(
            QtCore.Qt.Qt.AlignmentFlag.AlignRight
            | QtCore.Qt.Qt.AlignmentFlag.AlignTrailing
            | QtCore.Qt.Qt.AlignmentFlag.AlignVCenter
        )
        self.pixnumY_doubleSpinBox.setDecimals(0)
        self.pixnumY_doubleSpinBox.setObjectName("pixnumY_doubleSpinBox")
        self.coordY_doubleSpinBox = QtWidgets.QDoubleSpinBox(self.mainimagedata_groupBox)
        self.coordY_doubleSpinBox.setGeometry(QtCore.QRect(100, 55, 67, 24))
        self.coordY_doubleSpinBox.setAlignment(
            QtCore.Qt.Qt.AlignmentFlag.AlignRight
            | QtCore.Qt.Qt.AlignmentFlag.AlignTrailing
            | QtCore.Qt.Qt.AlignmentFlag.AlignVCenter
        )
        self.coordY_doubleSpinBox.setDecimals(3)
        self.coordY_doubleSpinBox.setObjectName("coordY_doubleSpinBox")
        self.pixnumZ_doubleSpinBox = QtWidgets.QDoubleSpinBox(self.mainimagedata_groupBox)
        self.pixnumZ_doubleSpinBox.setGeometry(QtCore.QRect(30, 80, 67, 24))
        self.pixnumZ_doubleSpinBox.setAlignment(
            QtCore.Qt.Qt.AlignmentFlag.AlignRight
            | QtCore.Qt.Qt.AlignmentFlag.AlignTrailing
            | QtCore.Qt.Qt.AlignmentFlag.AlignVCenter
        )
        self.pixnumZ_doubleSpinBox.setDecimals(0)
        self.pixnumZ_doubleSpinBox.setObjectName("pixnumZ_doubleSpinBox")
        self.coordZ_doubleSpinBox = QtWidgets.QDoubleSpinBox(self.mainimagedata_groupBox)
        self.coordZ_doubleSpinBox.setGeometry(QtCore.QRect(100, 80, 67, 24))
        self.coordZ_doubleSpinBox.setAlignment(
            QtCore.Qt.Qt.AlignmentFlag.AlignRight
            | QtCore.Qt.Qt.AlignmentFlag.AlignTrailing
            | QtCore.Qt.Qt.AlignmentFlag.AlignVCenter
        )
        self.coordZ_doubleSpinBox.setDecimals(3)
        self.coordZ_doubleSpinBox.setObjectName("coordZ_doubleSpinBox")
        self.label_6 = QtWidgets.QLabel(self.mainimagedata_groupBox)
        self.label_6.setGeometry(QtCore.QRect(10, 60, 16, 16))
        self.label_6.setObjectName("label_6")
        self.label_8 = QtWidgets.QLabel(self.mainimagedata_groupBox)
        self.label_8.setGeometry(QtCore.QRect(10, 85, 16, 16))
        self.label_8.setObjectName("label_8")
        self.coordT_doubleSpinBox = QtWidgets.QDoubleSpinBox(self.mainimagedata_groupBox)
        self.coordT_doubleSpinBox.setGeometry(QtCore.QRect(100, 130, 67, 24))
        self.coordT_doubleSpinBox.setAlignment(
            QtCore.Qt.Qt.AlignmentFlag.AlignRight
            | QtCore.Qt.Qt.AlignmentFlag.AlignTrailing
            | QtCore.Qt.Qt.AlignmentFlag.AlignVCenter
        )
        self.coordT_doubleSpinBox.setDecimals(3)
        self.coordT_doubleSpinBox.setObjectName("coordT_doubleSpinBox")
        self.pixnumT_doubleSpinBox = QtWidgets.QDoubleSpinBox(self.mainimagedata_groupBox)
        self.pixnumT_doubleSpinBox.setGeometry(QtCore.QRect(30, 130, 67, 24))
        self.pixnumT_doubleSpinBox.setAlignment(
            QtCore.Qt.Qt.AlignmentFlag.AlignRight
            | QtCore.Qt.Qt.AlignmentFlag.AlignTrailing
            | QtCore.Qt.Qt.AlignmentFlag.AlignVCenter
        )
        self.pixnumT_doubleSpinBox.setDecimals(0)
        self.pixnumT_doubleSpinBox.setObjectName("pixnumT_doubleSpinBox")
        self.label_3 = QtWidgets.QLabel(self.mainimagedata_groupBox)
        self.label_3.setGeometry(QtCore.QRect(10, 135, 16, 16))
        self.label_3.setObjectName("label_3")
        self.TimeSlider = QtWidgets.QSlider(self.mainimagedata_groupBox)
        self.TimeSlider.setGeometry(QtCore.QRect(10, 195, 151, 22))
        self.TimeSlider.setOrientation(QtCore.Qt.Qt.Orientation.Horizontal)
        self.TimeSlider.setObjectName("TimeSlider")
        self.label_10 = QtWidgets.QLabel(self.mainimagedata_groupBox)
        self.label_10.setGeometry(QtCore.QRect(30, 120, 56, 13))
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        self.label_10.setFont(font)
        self.label_10.setObjectName("label_10")
        self.label_11 = QtWidgets.QLabel(self.mainimagedata_groupBox)
        self.label_11.setGeometry(QtCore.QRect(100, 120, 56, 13))
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        self.label_11.setFont(font)
        self.label_11.setObjectName("label_11")
        self.runMovieButton = QtWidgets.QPushButton(self.mainimagedata_groupBox)
        self.runMovieButton.setGeometry(QtCore.QRect(20, 160, 151, 32))
        self.runMovieButton.setObjectName("runMovieButton")
        self.regressor_groupBox = QtWidgets.QGroupBox(self.centralwidget)
        self.regressor_groupBox.setGeometry(QtCore.QRect(10, 595, 1571, 181))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.regressor_groupBox.sizePolicy().hasHeightForWidth())
        self.regressor_groupBox.setSizePolicy(sizePolicy)
        self.regressor_groupBox.setMinimumSize(QtCore.QSize(100, 100))
        self.regressor_groupBox.setMaximumSize(QtCore.QSize(1590, 1000))
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        self.regressor_groupBox.setFont(font)
        self.regressor_groupBox.setObjectName("regressor_groupBox")
        self.regressorspectrum_graphicsView = GraphicsLayoutWidget(self.regressor_groupBox)
        self.regressorspectrum_graphicsView.setGeometry(QtCore.QRect(730, 25, 710, 131))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.regressorspectrum_graphicsView.sizePolicy().hasHeightForWidth()
        )
        self.regressorspectrum_graphicsView.setSizePolicy(sizePolicy)
        self.regressorspectrum_graphicsView.setMinimumSize(QtCore.QSize(710, 100))
        self.regressorspectrum_graphicsView.setMaximumSize(QtCore.QSize(710, 1000))
        self.regressorspectrum_graphicsView.setSizeIncrement(QtCore.QSize(1, 1))
        self.regressorspectrum_graphicsView.setObjectName("regressorspectrum_graphicsView")
        self.regressortimecourse_graphicsView = GraphicsLayoutWidget(self.regressor_groupBox)
        self.regressortimecourse_graphicsView.setEnabled(True)
        self.regressortimecourse_graphicsView.setGeometry(QtCore.QRect(10, 23, 710, 131))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.regressortimecourse_graphicsView.sizePolicy().hasHeightForWidth()
        )
        self.regressortimecourse_graphicsView.setSizePolicy(sizePolicy)
        self.regressortimecourse_graphicsView.setMinimumSize(QtCore.QSize(710, 100))
        self.regressortimecourse_graphicsView.setMaximumSize(QtCore.QSize(710, 1000))
        self.regressortimecourse_graphicsView.setSizeIncrement(QtCore.QSize(1, 1))
        self.regressortimecourse_graphicsView.setObjectName("regressortimecourse_graphicsView")
        self.processingstage_groupBox = QtWidgets.QGroupBox(self.regressor_groupBox)
        self.processingstage_groupBox.setGeometry(QtCore.QRect(1450, 5, 120, 151))
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        self.processingstage_groupBox.setFont(font)
        self.processingstage_groupBox.setObjectName("processingstage_groupBox")
        self.verticalLayoutWidget = QtWidgets.QWidget(self.processingstage_groupBox)
        self.verticalLayoutWidget.setGeometry(QtCore.QRect(10, 25, 101, 122))
        self.verticalLayoutWidget.setObjectName("verticalLayoutWidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.verticalLayoutWidget)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.prefilt_radioButton = QtWidgets.QRadioButton(self.verticalLayoutWidget)
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        self.prefilt_radioButton.setFont(font)
        self.prefilt_radioButton.setChecked(True)
        self.prefilt_radioButton.setObjectName("prefilt_radioButton")
        self.verticalLayout.addWidget(self.prefilt_radioButton)
        self.postfilt_radioButton = QtWidgets.QRadioButton(self.verticalLayoutWidget)
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        self.postfilt_radioButton.setFont(font)
        self.postfilt_radioButton.setChecked(False)
        self.postfilt_radioButton.setObjectName("postfilt_radioButton")
        self.verticalLayout.addWidget(self.postfilt_radioButton)
        self.pass1_radioButton = QtWidgets.QRadioButton(self.verticalLayoutWidget)
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        self.pass1_radioButton.setFont(font)
        self.pass1_radioButton.setObjectName("pass1_radioButton")
        self.verticalLayout.addWidget(self.pass1_radioButton)
        self.pass2_radioButton = QtWidgets.QRadioButton(self.verticalLayoutWidget)
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        self.pass2_radioButton.setFont(font)
        self.pass2_radioButton.setObjectName("pass2_radioButton")
        self.verticalLayout.addWidget(self.pass2_radioButton)
        self.pass3_radioButton = QtWidgets.QRadioButton(self.verticalLayoutWidget)
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        self.pass3_radioButton.setFont(font)
        self.pass3_radioButton.setObjectName("pass3_radioButton")
        self.verticalLayout.addWidget(self.pass3_radioButton)
        self.pass4_radioButton = QtWidgets.QRadioButton(self.verticalLayoutWidget)
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        self.pass4_radioButton.setFont(font)
        self.pass4_radioButton.setObjectName("pass4_radioButton")
        self.verticalLayout.addWidget(self.pass4_radioButton)
        spacerItem = QtWidgets.QSpacerItem(
            20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding
        )
        self.verticalLayout.addItem(spacerItem)
        self.label_5 = QtWidgets.QLabel(self.regressor_groupBox)
        self.label_5.setGeometry(QtCore.QRect(270, 155, 161, 16))
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        self.label_5.setFont(font)
        self.label_5.setAlignment(QtCore.Qt.Qt.AlignmentFlag.AlignCenter)
        self.label_5.setObjectName("label_5")
        self.label_12 = QtWidgets.QLabel(self.regressor_groupBox)
        self.label_12.setGeometry(QtCore.QRect(1040, 155, 131, 16))
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        self.label_12.setFont(font)
        self.label_12.setAlignment(QtCore.Qt.Qt.AlignmentFlag.AlignCenter)
        self.label_12.setObjectName("label_12")
        self.logOutput_groupBox = QtWidgets.QGroupBox(self.centralwidget)
        self.logOutput_groupBox.setGeometry(QtCore.QRect(1200, 7, 381, 221))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.logOutput_groupBox.sizePolicy().hasHeightForWidth())
        self.logOutput_groupBox.setSizePolicy(sizePolicy)
        self.logOutput_groupBox.setMinimumSize(QtCore.QSize(100, 100))
        self.logOutput_groupBox.setMaximumSize(QtCore.QSize(700, 500))
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        self.logOutput_groupBox.setFont(font)
        self.logOutput_groupBox.setObjectName("logOutput_groupBox")
        self.logOutput = QtWidgets.QTextEdit(self.logOutput_groupBox)
        self.logOutput.setGeometry(QtCore.QRect(10, 25, 361, 186))
        font = QtGui.QFont()
        font.setFamily("Courier New")
        font.setPointSize(11)
        font.setBold(True)
        self.logOutput.setFont(font)
        self.logOutput.setObjectName("logOutput")
        self.simfunc_groupBox = QtWidgets.QGroupBox(self.centralwidget)
        self.simfunc_groupBox.setGeometry(QtCore.QRect(1010, 407, 571, 186))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.simfunc_groupBox.sizePolicy().hasHeightForWidth())
        self.simfunc_groupBox.setSizePolicy(sizePolicy)
        self.simfunc_groupBox.setMinimumSize(QtCore.QSize(100, 100))
        self.simfunc_groupBox.setMaximumSize(QtCore.QSize(800, 1000))
        font = QtGui.QFont()
        font.setFamily("Arial Narrow")
        self.simfunc_groupBox.setFont(font)
        self.simfunc_groupBox.setObjectName("simfunc_groupBox")
        self.simfunc_graphicsView = GraphicsLayoutWidget(self.simfunc_groupBox)
        self.simfunc_graphicsView.setGeometry(QtCore.QRect(10, 25, 551, 151))
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.MinimumExpanding
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.simfunc_graphicsView.sizePolicy().hasHeightForWidth())
        self.simfunc_graphicsView.setSizePolicy(sizePolicy)
        self.simfunc_graphicsView.setMinimumSize(QtCore.QSize(0, 100))
        self.simfunc_graphicsView.setMaximumSize(QtCore.QSize(1000, 1000))
        self.simfunc_graphicsView.setObjectName("simfunc_graphicsView")
        self.overlays_groupBox.raise_()
        self.imageData_groupBox.raise_()
        self.histogram_groupBox.raise_()
        self.mainimagedata_groupBox.raise_()
        self.regressor_groupBox.raise_()
        self.logOutput_groupBox.raise_()
        self.simfunc_groupBox.raise_()
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1600, 24))
        self.menubar.setObjectName("menubar")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.actionQuit = QtWidgets.QAction(MainWindow)
        self.actionQuit.setObjectName("actionQuit")

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.imageData_groupBox.setTitle(_translate("MainWindow", "Image Data"))
        self.resetDispLimits_Button.setToolTip(
            _translate(
                "MainWindow",
                "Set the display range of the current map to the absolute minimum and maximum",
            )
        )
        self.resetDispLimits_Button.setText(_translate("MainWindow", "Full Range"))
        self.resetDispSmart_Button.setToolTip(
            _translate(
                "MainWindow",
                "Set the display range of the current map to the robust minimum and maximum",
            )
        )
        self.resetDispSmart_Button.setText(_translate("MainWindow", "Smart"))
        self.saveDisp_Button.setToolTip(
            _translate("MainWindow", "Save jpegs of the currently displayed map")
        )
        self.saveDisp_Button.setText(_translate("MainWindow", "Save"))
        self.setMask_Button.setToolTip(
            _translate(
                "MainWindow", "Right click to select the mask used to display the current map"
            )
        )
        self.setMask_Button.setText(_translate("MainWindow", "No mask"))
        self.transparency_checkBox.setToolTip(
            _translate(
                "MainWindow",
                "With transparency on, map values outside the range limits are not displayed.",
            )
        )
        self.transparency_checkBox.setText(_translate("MainWindow", "Transparency"))
        self.label_2.setText(_translate("MainWindow", "Colormap"))
        self.gray_radioButton.setText(_translate("MainWindow", "Gray"))
        self.thermal_radioButton.setText(_translate("MainWindow", "Thermal"))
        self.plasma_radioButton.setText(_translate("MainWindow", "Plasma"))
        self.rainbow_radioButton.setText(_translate("MainWindow", "Rainbow"))
        self.viridis_radioButton.setText(_translate("MainWindow", "Viridis"))
        self.turbo_radioButton.setText(_translate("MainWindow", "Turbo"))
        self.summary_Label.setText(_translate("MainWindow", "Region summaries"))
        self.raw_radioButton.setText(_translate("MainWindow", "Raw data"))
        self.mean_radioButton.setText(_translate("MainWindow", "Mean"))
        self.median_radioButton.setText(_translate("MainWindow", "Median"))
        self.CoV_radioButton.setText(_translate("MainWindow", "CoV"))
        self.std_radioButton.setText(_translate("MainWindow", "Std dev"))
        self.MAD_radioButton.setText(_translate("MainWindow", "MAD"))
        self.overlays_groupBox.setTitle(_translate("MainWindow", "Overlays"))
        self.overlay_radioButton_05.setText(_translate("MainWindow", "Mask"))
        self.overlay_radioButton_01.setText(_translate("MainWindow", "Lag times"))
        self.overlay_radioButton_02.setText(_translate("MainWindow", "Correlation strength"))
        self.overlay_radioButton_03.setText(_translate("MainWindow", "MTT"))
        self.overlay_radioButton_04.setText(_translate("MainWindow", "R2"))
        self.corrout_radioButton.setText(_translate("MainWindow", "Correlation function"))
        self.overlay_radioButton_06.setText(_translate("MainWindow", "Mask"))
        self.overlay_radioButton_07.setText(_translate("MainWindow", "Mask"))
        self.overlay_radioButton_08.setText(_translate("MainWindow", "Mask"))
        self.overlay_radioButton_09.setText(_translate("MainWindow", "Lag times"))
        self.overlay_radioButton_10.setText(_translate("MainWindow", "Lag times"))
        self.overlay_radioButton_11.setText(_translate("MainWindow", "Lag times"))
        self.overlay_radioButton_12.setText(_translate("MainWindow", "Lag times"))
        self.overlay_radioButton_13.setText(_translate("MainWindow", "Lag times"))
        self.overlay_radioButton_14.setText(_translate("MainWindow", "Lag times"))
        self.overlay_radioButton_15.setText(_translate("MainWindow", "Lag times"))
        self.overlay_radioButton_16.setText(_translate("MainWindow", "Lag times"))
        self.histogram_groupBox.setTitle(_translate("MainWindow", "Overlay Histogram"))
        self.mainimagedata_groupBox.setTitle(_translate("MainWindow", "Location"))
        self.label_4.setText(_translate("MainWindow", "X"))
        self.label_7.setText(_translate("MainWindow", "Voxel"))
        self.label_9.setText(_translate("MainWindow", "mm"))
        self.label_6.setText(_translate("MainWindow", "Y"))
        self.label_8.setText(_translate("MainWindow", "Z"))
        self.label_3.setText(_translate("MainWindow", "T"))
        self.label_10.setText(_translate("MainWindow", "Timepoint"))
        self.label_11.setText(_translate("MainWindow", "seconds"))
        self.runMovieButton.setText(_translate("MainWindow", "Start Movie"))
        self.regressor_groupBox.setTitle(_translate("MainWindow", "Probe regressor"))
        self.processingstage_groupBox.setTitle(_translate("MainWindow", "Processing stage"))
        self.prefilt_radioButton.setText(_translate("MainWindow", "Prefilt"))
        self.postfilt_radioButton.setText(_translate("MainWindow", "Postfilt"))
        self.pass1_radioButton.setText(_translate("MainWindow", "Pass 1"))
        self.pass2_radioButton.setText(_translate("MainWindow", "Pass 2"))
        self.pass3_radioButton.setText(_translate("MainWindow", "Pass 3"))
        self.pass4_radioButton.setText(_translate("MainWindow", "Pass 4"))
        self.label_5.setText(_translate("MainWindow", "Timecourse (seconds)"))
        self.label_12.setText(_translate("MainWindow", "Power spectrum (Hz)"))
        self.logOutput_groupBox.setTitle(_translate("MainWindow", "Information"))
        self.simfunc_groupBox.setTitle(_translate("MainWindow", "Correlation Function"))
        self.actionQuit.setText(_translate("MainWindow", "Quit"))


from pyqtgraph import GraphicsLayoutWidget, GraphicsView
